/* extract_properties interprets the document tree and creates a property store
 *
 * Author: Bert Bos
 * Created: 16 August 2018
 *
 * Copyright © 2018 World Wide Web Consortium, (Massachusetts
 * Institute of Technology, European Research Consortium for
 * Informatics and Mathematics, Keio University, Beihang). All Rights
 * Reserved. See
 * http://www.w3.org/Consortium/Legal/copyright-software
 */

#include "config.h"
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <err.h>
#include "export.h"
#include "heap.e"
#include "types.e"
#include "propstore.e"
#include "tree.e"

#define XMLNS       "xml:"
#define LANG        "xml:lang"
#define XMPMETA     "adobe:ns:meta/xmpmeta"
#define RDFNS       "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
#define RDF         "http://www.w3.org/1999/02/22-rdf-syntax-ns#RDF"
#define DESCRIPTION "http://www.w3.org/1999/02/22-rdf-syntax-ns#Description"
#define ABOUT       "http://www.w3.org/1999/02/22-rdf-syntax-ns#about"
#define BAG         "http://www.w3.org/1999/02/22-rdf-syntax-ns#Bag"
#define ALT         "http://www.w3.org/1999/02/22-rdf-syntax-ns#Alt"
#define SEQ         "http://www.w3.org/1999/02/22-rdf-syntax-ns#Seq"
#define LI          "http://www.w3.org/1999/02/22-rdf-syntax-ns#li"
#define RESOURCE    "http://www.w3.org/1999/02/22-rdf-syntax-ns#resource"
#define PARSETYPE   "http://www.w3.org/1999/02/22-rdf-syntax-ns#parseType"


static void extract_from_description(Tree t, Store *store, string lang,
				     string *resource);
static void extract_from_element(Tree t, Store *store, string lang);


/* parse_attributes -- parse and remove special attributes */
static void parse_attributes(pairlist *attribs, bool with_about,
			     string *resource, bool *structure, string *lang)
{
  pairlist *a = attribs;

  *resource = NULL;
  *structure = false;

  while (*a) {
    if (eq((*a)->name, LANG)) {
      *lang = newstring((*a)->value);
      *a = (*a)->next;		/* Remove attribute */
    } else if (eq((*a)->name, RESOURCE)) {
      *resource = newstring((*a)->value);
      *a = (*a)->next;		/* Remove attribute */
    } else if (eq((*a)->name, ABOUT)) {
      if (!with_about) warn("Attribute rdf:about ignored\n");
      else if ((*a)->value[0]) warn("Ignored rdf:about=\"%s\"\n", (*a)->value);
      *a = (*a)->next;		/* Remove attribute */
    } else if (eq((*a)->name, PARSETYPE) && eq((*a)->value, "Resource")) {
      *structure = true;
      *a = (*a)->next;		/* Remove attribute */
    } else if ((!eq((*a)->name, VALUE) && hasprefix((*a)->name, RDFNS)) ||
	       hasprefix((*a)->name, XMLNS)) {
      warnx("Attribute %s ignored", (*a)->name);
      *a = (*a)->next;		/* Remove attribute */
    } else
      a = &(*a)->next;		/* Leave this attribute. Go to the next. */
    }
}


/* extract_from_alt_bag_or_seq -- extract values from an Alt, Seq or Bag */
static void extract_from_alt_bag_or_seq(Tree t, Store *store, string lang)
{
  Tree h;

  assert(t->tp == Element &&
	 (eq(t->name, ALT) || eq(t->name, BAG) || eq(t->name, SEQ)));

  if (t->attribs) warnx("Ignored attribute(s) on %s", t->name);
  for (h = t->children; h; h = h->sister) {
    assert(h->tp == Text || h->tp == Element);
    if (h->tp == Element && eq(h->name, LI))
      extract_from_element(h, store, lang);
    else if (h->tp == Element)
      warnx("Ignored element %s outside <rdf:li>", h->name);
    else if (!only_space(h->text))
      warnx("Ignored text outside <rdf:li>");
  }
}


/* get_element_content -- get the text content of an element */
static string get_element_content(Tree t)
{
  string s;
  Tree h;

  s = newstring("");
  for (h = t->children; h; h = h->sister)
    if (h->tp == Text) s = strapp(&s, h->text, NULL);
    else warnx("Sub-property %s ignored in property %s", h->name, t->name);
  return s;
}


/* extract_from_element -- extract values from a property element or <rdf:li> */
static void extract_from_element(Tree t, Store *store, string lang)
{
  string resource, s, name;
  PropertyType type;
  bool structure;
  pairlist a;
  Store sub;
  Tree h;

  assert(t->tp == Element);
  assert(lang);

  parse_attributes(&t->attribs, false, &resource, &structure, &lang);

  sub = store_create();

  /* Handle attributes */
  for (a = t->attribs; a; a = a->next)
    store_add(&sub, a->name, Simple, a->value, NULL, lang);

  /* Handle content */
  s = newstring("");
  type = structure ? Structure : Simple;
  for (h = t->children; h; h = h->sister) {
    assert(t->tp == Element || t->tp == Text);
    if (h->tp == Text) {
      strapp(&s, h->text, NULL);
    } else if (eq(h->name, DESCRIPTION)) {
      if (type != Simple) warnx("Unexpected rdf:Description in %s", t->name);
      type = Structure;
      extract_from_description(h, &sub, lang, &resource);
    } else if (eq(h->name, BAG)) {
      if (type != Simple) warnx("Unexpected rdf:Bag in %s", t->name);
      type = Bag;
      extract_from_alt_bag_or_seq(h, &sub, lang);
    } else if (eq(h->name, ALT)) {
      if (type != Simple) warnx("Unexpected rdf:Alt inside %s", t->name);
      type = Alt;
      extract_from_alt_bag_or_seq(h, &sub, lang);
    } else if (eq(h->name, SEQ)) {
      if (type != Simple) warnx("Unexpected rdf:Seq inside %s", t->name);
      type = Seq;
      extract_from_alt_bag_or_seq(h, &sub, lang);
    } else {
      if (!structure)
	warnx("Unexpected element %s inside %s", h->name, t->name);
      else if (type != Structure && type != Simple)
	warnx("Found both an array and other elements inside %s", t->name);
      type = Structure;
      extract_from_element(h, &sub, lang);
    }
  }

  if (type != Simple && !only_space(s))
    warnx("Found both text content and elements in %s", t->name);

  if (resource) {
    if (type == Alt || type == Bag || type == Seq)
      warnx("Found both rdf:resource and an array in %s", t->name);
    if (!only_space(s))
      warnx("Found both rdf:resource and text content in %s", t->name);
    type = Resource;
    dispose(s); s = resource;
  } else if (type == Simple && !store_is_empty(sub)) {
    type = Structure;
  }

  /* The name is NULL if we're inside an Alt, Bag or Seq */
  name = eq(t->name, LI) ? NULL : t->name;

  store_add(store, name, type, s, sub, lang);
}


/* extract_from_description -- extract properties from <rdf:Description> */
static void extract_from_description(Tree t, Store *store, string lang,
				     string *resource)
{
  bool structure;
  pairlist a;
  Tree h;

  assert(t->tp == Element && eq(t->name, DESCRIPTION));
  assert(lang);

  parse_attributes(&t->attribs, true, resource, &structure, &lang);
  if (structure) warnx("rdf:parseType not allowed on <rdf:Description>");
  for (a = t->attribs; a; a = a->next)
    store_add(store, a->name, Simple, a->value, NULL, lang);

  for (h = t->children; h; h = h->sister) {
    assert(h->tp == Element || h->tp == Text);
    if (h->tp == Text) {
      if (!only_space(h->text)) warnx("Text in <rdf:Description> ignored");
    } else if (eq(h->name, RESOURCE)) {
      if (*resource) warnx("Second rdf:resource in rdf:Description ignored");
      else *resource = get_element_content(h);
    } else {			/* Not rdf:resource */
      extract_from_element(h, store, lang);
    }
  }
}


/* extract_from_rdf -- extract properties from <rdf:RDF> and below */
static void extract_from_rdf(Tree t, Store *store, string lang)
{
  string resource;
  pairlist a;
  Tree h;

  assert(t->tp == Element && eq(t->name, RDF));

  /* Get xml:lang attribute, if any. */
  for (a = t->attribs; a; a = a->next)
    if (eq(a->name, LANG)) lang = newstring(a->value);
    else warnx("Extra attribute on <rdf:RDF> ignored: %s", a->name);

  for (h = t->children; h; h = h->sister)
    if (h->tp == Element && !eq(h->name, DESCRIPTION)) {
      warnx("Expected <rdf:Description> in <rdf:RDF>, but found %s",h->name);
    } else if (h->tp == Element) {
      extract_from_description(h, store, lang, &resource);
      if (resource) warnx("rdf:resource on top-level rdf:Description ignored");
    } else {
      assert(h->tp == Text);
      if (!only_space(h->text)) warnx("Text in <rdf:RDF> ignored");
    }
}


/* extract_from_xmpmeta -- extract properties from <x:xmpmeta> and below */
static void extract_from_xmpmeta(Tree t, Store *store)
{
  string lang = "", resource;
  bool structure;
  Tree h;

  assert(t->tp == Element && eq(t->name, XMPMETA));

  parse_attributes(&t->attribs, false, &resource, &structure, &lang);
  if (resource) warnx("Attribute rdf:resource on <x:xmpmeta> ignored\n");
  if (structure) warnx("Attribute rdf:parseType on <x:xmpmeta> ignored\n");
  /* Ignore all other attributes except language */

  /* TODO: Check that there is only one <rdf:RDF> */
  for (h = t->children; h; h = h->sister)
    if (h->tp == Element && eq(h->name, RDF))
      extract_from_rdf(h, store, lang);
    else if (h->tp == Element)
      warnx("Only <rdf:RDF> allowed inside <x:xmpmeta>. Ignored %s", h->name);
    else {
      assert(h->tp == Text);
      if (!only_space(h->text)) warnx("Text in <x:xmpmeta> ignored");
    }
}


/* extract_from_tree -- walk the document, storing properties in the store */
EXPORT void extract_from_tree(Tree t, Store *store)
{
  Tree h;

  if (t->tp == Root)
    for (h = t->children; h; h = h->sister) extract_from_tree(h, store);
  else if (t->tp == Element && eq(t->name, XMPMETA))
    extract_from_xmpmeta(t, store);
  else if (t->tp == Element && eq(t->name, RDF))
    extract_from_rdf(t, store, "");
  else if (t->tp == Element)
    warnx("Top level should be <x:xmpmeta> or <rdf:RDF>. Ignored %s", t->name);
  else
    assert(t->tp == Text);
}


/* extract_properties -- walk the document, storing properties in the store */
EXPORT void extract_properties(Tree t, Store *store)
{
  *store = store_create();
  extract_from_tree(t, store);
}
  
