/* A SymbolTable implements a symbol table with nested scopes to look
 * up one string (a namespace prefix) and translate it to another (a
 * namespace URI). Neither may be NULL. The empty prefix ("")
 * represents the default namespace.
 *
 * Implementation:
 *
 * This implementation uses simple linear search to find a given
 * string and is thus only suitable for small numbers of strings.
 *
 * The symbol table is implemented as a linked list, acting as a
 * stack. Callers hold a pointer to what is, for them, the top of the
 * stack. They add new entries at this top. There is no function to
 * explicitly close a scope and memory is not recovered. (The stack is
 * thus in fact a cactus stack: popped items stay on as old branches.)
 *
 * Author: Bert Bos <bert@w3.org>
 * Created: 15 August 2018
 *
 * Copyright © 2018 World Wide Web Consortium, (Massachusetts
 * Institute of Technology, European Research Consortium for
 * Informatics and Mathematics, Keio University, Beihang). All Rights
 * Reserved. See http://www.w3.org/Consortium/Legal/copyright-software
*/

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "export.h"
#include "heap.e"
#include "types.e"

EXPORT typedef struct _SymbolTable *SymbolTable;

struct _SymbolTable {
  char *prefix;
  char *namespace;
  SymbolTable next;
};


/* symboltable_add -- add a string and its definition to a symbol table */
EXPORT void symboltable_add(SymbolTable *table, const char *prefix,
			    const char *namespace)
{
  SymbolTable h;

  assert(prefix != NULL);
  assert(namespace != NULL);
  new(h);
  h->prefix = newstring(prefix);
  h->namespace = newstring(namespace);
  h->next = *table;
  *table = h;
}


/* symboltable_create -- return a new symbol table with only the XML namespace */
EXPORT SymbolTable symboltable_create(void)
{
  SymbolTable table = NULL;

  symboltable_add(&table, "xml", "xml:");
  return table;
}


/* symboltable_lookup -- return the string that corresponds to prefix, or NULL */
EXPORT string symboltable_lookup(const SymbolTable table, const char *prefix)
{
  if (table == NULL) return NULL;
  else if (eq(prefix, table->prefix)) return table->namespace;
  else return symboltable_lookup(table->next, prefix);
}


/* globalname -- look up prefix, return concatenated namespace and name */
EXPORT char* globalname(const SymbolTable table, const char *prefix,
			const char *name)
{
  char *namespace;

  assert(prefix);
  assert(name);
  namespace = symboltable_lookup(table, prefix);
  if (!namespace) return NULL;
  else return strapp(NULL, namespace, name, NULL);
}

