/* print_xmp prints a property store on stdout as RDF/XML.
 *
 * Author: Bert Bos
 * Created: 16 August 2018
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "export.h"
#include "heap.e"
#include "types.e"
#include "propstore.e"
#include "xmlnames.e"


/* A few of the most common namespaces and their recommended prefixes */
static string well_known[][2] = {
  {"http://ns.adobe.com/camera-raw-settings/1.0/",	"crs"},
  {"http://ns.adobe.com/exif/1.0/",			"exif"},
  {"http://ns.adobe.com/exif/1.0/aux/",			"aux"},
  {"http://ns.adobe.com/photoshop/1.0/",		"photoshop"},
  {"http://ns.adobe.com/tiff/1.0/",			"tiff"},
  {"http://ns.adobe.com/xap/1.0/",			"xap"},
  {"http://purl.org/dc/elements/1.1/",			"dc"},
  {"http://www.w3.org/1999/02/22-rdf-syntax-ns#",	"rdf"},
  {"http://www.w3.org/2000/PhotoRDF/technical-1-0#",	"tc"},
  {NULL, NULL}};

static void print_cb(string name, PropertyType type, string val, Store sub,
		     string lang, void *data);



/* well_known_prefix -- return a prefix if the namespace is well_known */
static string well_known_prefix(const string ns)
{
  int i;

  for (i = 0; well_known[i][0]; i++)
    if (eq(ns, well_known[i][0])) return well_known[i][1];
  return NULL;
}


/* escape -- print a string with XML delimiters replaced by character entities */
static void escape(string s)
{
  /* No need to escape the single quote, all our strings use double quotes */
  for (; *s; s++)
    switch (*s) {
    case '&': printf("&amp;"); break;
    case '<': printf("&lt;"); break;
    case '>': printf("&gt;"); break;
    case '"': printf("&quot;"); break;
    default: putchar(*s);
    }
}


/* print_item_cb -- callback to print a value in a Bag, Alt or Seq */
static void print_item_cb(string name, PropertyType type, string val, Store sub,
			  string lang, void *data)
{
  int indent = *(int*)data;
  int indent2 = indent + 2;

  printf("%*s<rdf:li", indent, "");
  if (type == Simple) {
    if (*lang) printf(" xml:lang=\"%s\"", lang);
    printf(">");
    escape(val);
    printf("</rdf:li>\n");
  } else if (type == Resource) {
    printf(" rdf:resource=\"");
    escape(val);
    printf("\"/>\n");
  } else if (type == Bag) {
    printf(">\n%*s<rdf:Bag>\n", indent + 1, "");
    store_walk(sub, print_item_cb, &indent2);
    printf("%*s</rdf:Bag>\n%*s</rdf:li>\n", indent + 1, "", indent, "");
  } else if (type == Alt) {
    printf(">\n%*s<rdf:Alt>\n", indent + 1, "");
    store_walk(sub, print_item_cb, &indent2);
    printf("%*s</rdf:Alt>\n%*s</rdf:li>\n", indent + 1, "", indent, "");
  } else if (type == Seq) {
    printf(">\n%*s<rdf:Seq>\n", indent + 1, "");
    store_walk(sub, print_item_cb, &indent2);
    printf("%*s</rdf:Seq>\n%*s</rdf:li>\n", indent + 1, "", indent, "");
  } else {
    assert(type == Structure);
    printf(">\n%*s<rdf:Description>\n", indent + 1, "");
    store_walk(sub, print_cb, &indent2);
    printf("%*s</rdf:Description>\n%*s</rdf:li>\n", indent + 1, "", indent, "");
  }
}


/* print_cb -- callback to print a property and value */
static void print_cb(string property, PropertyType type, string val, Store sub,
		     string lang, void *data)
{
  int indent = *(int*)data, indent2 = indent + 2;
  string namespace, name, prefix, p;

  /* Split the property name in a namespace and a local name. */
  for (name = p = property; *p;)
    if (!is_name_char(p, &p)) name = p;
  namespace = newnstring(property, name - property);
  prefix = well_known_prefix(namespace);

  if (prefix) printf("%*s<%s:%s", indent, "", prefix, name);
  else printf("%*s<%s xmlns=\"%s\"", indent, "", name, namespace);

  switch (type) {

  case Simple:
    if (lang && *lang && !eq(property, VALUE)) printf(" xml:lang=\"%s\"", lang);
    printf(">");
    if (sub == NULL) {
      escape(val);
    } else {
      printf("\n%*s<rdf:Description>\n", indent + 1, "");
      printf("%*s<rdf:value>", indent + 2, "");
      escape(val);
      printf("</rdf:value>\n");
      store_walk(sub, print_cb, &indent2);
      printf("%*s</rdf:Description>\n%*s", indent + 1, "", indent, "");
    }
    if (prefix) printf("</%s:%s>\n", prefix, name);
    else printf("</%s>\n", name);
    break;

  case Resource:
    if (sub == NULL) {
      printf(" rdf:resource=\"");
      escape(val);
      printf("\"/>\n");
    } else {
      printf(">\n%*s<rdf:Description>\n", indent + 1, "");
      printf("%*s<rdf:value rdf:resource=\"", indent + 2, "");
      escape(val);
      printf("\"/>\n");
      store_walk(sub, print_cb, &indent2);
      printf("%*s</rdf:Description>\n", indent + 1, "");
      if (prefix) printf("%*s</%s:%s>\n", indent, "", prefix, name);
      else printf("%*s</%s>\n", indent, "", name);
    }
    break;

  case Structure:
    if (lang && *lang) printf(" xml:lang=\"%s\"", lang);
    printf(">\n%*s<rdf:Description>\n", indent + 1, "");
    store_walk(sub, print_cb, &indent2);
    printf("%*s</rdf:Description>\n", indent + 1, "");
    if (prefix) printf("%*s</%s:%s>\n", indent, "", prefix, name);
    else printf("%*s</%s>\n", indent, "", name);
    break;

  default:
    printf(">\n%*s<rdf:%s>\n", indent + 1, "",
	   type == Bag ? "Bag" : type == Alt ? "Alt" : "Seq");
    store_walk(sub, print_item_cb, &indent2);
    printf("%*s</rdf:%s>\n", indent + 1, "",
	   type == Bag ? "Bag" : type == Alt ? "Alt" : "Seq");
    if (prefix) printf("%*s</%s:%s>\n", indent, "", prefix, name);
    else printf("%*s</%s>\n", indent, "", name);
  }

  dispose(namespace);
}


/* print_xmp -- serialize the store as RDF/XML */
EXPORT void print_xmp(Store store)
{
  int i, indent = 2;

  printf("<?xpacket begin='\357\273\277' id='W5M0MpCehiHzreSzNTczkc9d'?>\n");
  printf("<rdf:RDF");
  for (i = 0; well_known[i][0]; i++)
    printf("\n  xmlns:%s=\"%s\"", well_known[i][1], well_known[i][0]);
  printf(">\n");
  printf(" <rdf:Description rdf:about=\"\">\n");

  store_walk(store, print_cb, &indent);

  printf(" </rdf:Description>\n");
  printf("</rdf:RDF>\n");
  printf("<?xpacket end='w'?>\n");
}


