/* A Store is an implementation of a property/value store, like an RDF
 * triple store for a single, implicit subject. Each property has a
 * name and a value, and the value can be of one of the following
 * types:
 *
 * Simple: the value is a text string.
 * Resource: the value is a string representing a URI.
 * Alt: the value is an array of values that are each other's alternatives.
 * Bag: the value is an unordered array of values.
 * Seq: the value is an ordered array of values.
 * Structure: the value is a set of named fields, each with a value.
 *
 * A value of type Simple has an associated language (a string).
 *
 * The values of an Alt, Bag, Seq or Structure are themselves
 * Stores. According to RDF, the values of an Alt, Bag and Seq have
 * names "rdf:_1", "rdf:_2", "rdf:_3", etc., but here they are left
 * implicit: their names are NULL.
 *
 * A Simple property is a "qualified" property if it has qualifiers
 * (sub-properties, i.e., sub != NULL)
 *
 * Each property name in a Store is unique. Setting a value on an
 * existing property automatically converts the property to a Bag
 * containing both the existing and the new value.
 *
 * Functions:
 *
 * store_create
 *
 *    Creates a new, empty store.
 *
 * store_is_empty
 *
 *    True if the store is empty.
 *
 * store_add
 *
 *    Add a property/value pair to the store.
 *
 * store_delete
 *
 *    Removes all matching properties from the store. A Simple
 *    property is removed if both the property name and the language
 *    match. A Resource or a Structure is removed if its name matches.
 *    If a property is an array (Alt, Bag or Seq) whose name matches,
 *    then all the array values are removed except for Simple values
 *    with a non-matching language. If there are none, i.e., the array
 *    ends up empty, then the whole property is removed.
 *
 * store_walk
 *
 *    Iterate over a store in alphabetical order of property names (or
 *    the order in which entries were added, in case of an Alt, Bag or
 *    Seq) and call a callback funtion for each entry. The function
 *    should not alter any of the values it is passed. The caller may
 *    pass in a pointer to arbitrary data that is passed on to the
 *    callback.
 *
 * store_lookup
 *
 *    Like store_walk, but only calls the callback for properties
 *    matching a given name and a given language. The name and the
 *    language are shell wildcard patterns: They may contain *, ? and
 *    []. (The patterns are defined by fnmatch(3).)
 *
 * store_simplify
 *
 *    Remove redundant values. A value is redundant if it has the same
 *    property name, same value and same language as another, or it
 *    has the same property name and value as another that has no
 *    language or a language of x-default. A Structure without any
 *    fields is also removed. And an array (Seq, Alt or Bag) with only
 *    one item is simplified to a Simple property.
 *
 * Implementation:
 *
 * A Store is an (unbalanced) binary tree sorted on the property name.
 *
 * Author: Bert Bos <bert@w3.org>
 * Created: 15 August 2018
 *
 * Copyright © 2018 World Wide Web Consortium, (Massachusetts
 * Institute of Technology, European Research Consortium for
 * Informatics and Mathematics, Keio University, Beihang). All Rights
 * Reserved. See http://www.w3.org/Consortium/Legal/copyright-software
*/

#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <fnmatch.h>
#include "export.h"
#include "heap.e"
#include "types.e"

#define VALUE    "http://www.w3.org/1999/02/22-rdf-syntax-ns#value"
EXPORTDEF(VALUE)

EXPORT typedef struct _Store *Store;
EXPORT typedef enum {Simple, Resource, Alt, Bag, Seq, Structure} PropertyType;
EXPORT typedef void (StoreCallback)(string property, PropertyType type,
				    string value, Store structure, string lang,
				    void *data);

struct _Store {			/* A sorted, unbalanced tree */
  string name;			/* Name including namespace */
  PropertyType type;
  string value;			/* For type Simple and Resource */
  string lang;			/* Only for type Simple */
  struct _Store *sub;		/* Array items, fields or qualifiers */
  struct _Store *left, *right;	/* Left and right sub-trees */
};


/* store_create -- return a new, empty store */
EXPORT Store store_create(void)
{
  return NULL;
}


/* store_is_empty -- return true if the store is empty */
EXPORT bool store_is_empty(Store store)
{
  return store == NULL;
}


/* match -- check if a property matches the given name pattern and language */
static bool match(Store store, const string pattern, const string lang)
{
  assert(store);
  assert(store->type != Simple || store->lang);
  assert(store->name == NULL || pattern);
  assert(lang);

  /* If we're in an Alt, Bag or Seq, the name is NULL and the pattern
     is ignored. If the property isn't Simple, the lang is ignored. */
  return
    (store->name == NULL || fnmatch(pattern, store->name, 0) == 0) &&
    (store->type != Simple || fnmatch(lang, store->lang, 0) == 0);
}


/* delete_node -- delete a node in a tree, return true */
static bool delete_node(Store *s)
{
  Store p;

  if ((*s)->left == NULL) { /* Left branch is empty, move right branch up */
    *s = (*s)->right;
  } else if ((*s)->right == NULL) { /* Right is empty, move left branch up */
    *s = (*s)->left;
  } else { /* Move the left brach to bottom left of the right branch */
    for (p = (*s)->right; p->left != NULL; p = p->left);
    p->left = (*s)->left;
    *s = (*s)->right;		/* Move right branch up */
  }
  return true;
}


/* store_delete -- delete matching properties from store, false if none found */
EXPORT bool store_delete(Store *s, const string pattern, const string lang)
{
  bool r = false;

  if (*s) {
    if ((*s)->left) r = store_delete(&(*s)->left, pattern, lang);
    if ((*s)->right) r = r || store_delete(&(*s)->right, pattern, lang);

    if (match(*s, pattern, lang)) {
      switch ((*s)->type) {
      case Simple: case Resource: case Structure:
	delete_node(s);
	r = true;
	break;
      case Alt: case Bag: case Seq:
	r = r || store_delete(&(*s)->sub,NULL,lang); /* Delete values by lang */
	if ((*s)->sub == NULL) delete_node(s), r = true; /* Nothing was left */
	break;
      default:
	assert(!"Cannot happen!");
      }
    }
  }
  return r;
}


/* compare -- compare names alphabetically, except for rdf:value */
static int compare(const string a, const string b)
{
  int r = strcmp(a, b);

  if (r == 0) return 0;			/* a and b are equal */
  else if (eq(a, VALUE)) return -1;	/* "rdf:value" always comes first */
  else if (eq(b, VALUE)) return 1;	/* "rdf:value" always comes first */
  else return r;
}


/* process_rdf_value -- convert a Structure to qualified property if needed */
static void process_rdf_value(Store store)
{
  Store h;

  if (store->type == Structure && store->sub) {
    /* Check if the Structure contains an rdf:value. (Must be leftmost.) */
    for (h = store->sub; h->left; h = h->left);
    if (eq(h->name, VALUE)) {
      /* Promote the language from the rdf:value to the property itself. */
      assert(store->lang == NULL);
      store->lang = h->lang;
      h->lang = NULL;

      /* Simplify if the Structure contains no actual qualifiers */
      if (h == store->sub && store->sub->right == NULL) {
	assert(store->left == NULL);
	store->type = store->sub->type;
	if (store->type == Simple || store->type == Resource)
	  store->value = store->sub->value;
	store->sub = store->sub->sub;
      }
    }
  }
}


/* store_add -- add a property/value pair to a store */
EXPORT void store_add(Store *store, const string property,
		      const PropertyType type, const string value,
		      const Store structure, const string lang)
{
  Store h;
  int r;

  /* Can only add a nameless property to an Alt, Bag or Seq: */
  assert(property || (!*store || !(*store)->name));
  assert((type != Simple && type != Resource) || value);

  if (*store == NULL) {
    assert(type != Simple || lang);
    /* No such property exists, create a new entry */
    new(*store);
    (*store)->name = newstring(property);
    (*store)->type = type;
    (*store)->lang = type == Simple && lang ? newstring(lang) : NULL;
    if (type == Simple || type == Resource) (*store)->value = newstring(value);
    (*store)->sub = structure;
    (*store)->left = (*store)->right = NULL;
    process_rdf_value(*store);	/* Handle rdf:value */
  } else if ((*store)->name == NULL) {
    /* We're adding to an Alt, Bag or Seq; just add to the right. */
    store_add(&(*store)->right, NULL, type, value, structure, lang);
  } else if ((r = compare(property, (*store)->name)) < 0) {
    /* Look in the left subtree (alphabetically earlier). */
    store_add(&(*store)->left, property, type, value, structure, lang);
  } else if (r > 0) {
    /* Look in the right subtree. */
    store_add(&(*store)->right, property, type, value, structure, lang);
  } else if ((*store)->type == Bag || (*store)->type == Seq ||
	     (*store)->type == Alt) {
    /* A property of that name already exists and its an Alt, Seq or Bag */
    store_add(&(*store)->sub, property, type, value, structure, lang);
  } else {
    /* Turn the existing property into a Bag */
    h = store_create();
    store_add(&h, NULL, (*store)->type, (*store)->value, (*store)->sub,
	      (*store)->lang);
    store_add(&h, NULL, type, value, structure, lang);
    (*store)->type = Bag;
    (*store)->sub = h;
  }
}


/* store_walk -- call callback for each entry in store, in alphabetical order */
EXPORT void store_walk(const Store s, StoreCallback fn, void *data)
{
  if (s) {
    store_walk(s->left, fn, data);
    fn(s->name, s->type, s->value, s->sub, s->lang, data);
    store_walk(s->right, fn, data);
  }
}


/* store_lookup -- walk the tree, calling callback for each matching property */
EXPORT void store_lookup(Store s, const string pattern, const string lang,
			 StoreCallback fn, void *data)
{
  if (s) {
    store_lookup(s->left, pattern, lang, fn, data);

    if (match(s, pattern, lang)) {
      if (s->type == Alt || s->type == Bag || s->type == Seq)
	/* Pass the property name instead of the pattern. It will not
	   be used to match, but it will be passed to the callback in
	   the last "else" below. */
	store_lookup(s->sub, s->name, lang, fn, data);
      else if (s->name)
	fn(s->name, s->type, s->value, s->sub, s->lang, data);
      else
	/* It's a value in an Alt, Bag or Seq. The pattern contains
	   the name of the property. */
	fn(pattern, s->type, s->value, s->sub, s->lang, data);
    }

  store_lookup(s->right, pattern, lang, fn, data);
  }
}


/* is_subsumed -- b contains a value that is the same or more generic than a */
static bool is_subsumed(Store a, Store b)
{
  if (!b) {
    return false;
  } else if (a == b) {
    return is_subsumed(a, b->right);
  } else if (a->type != b->type) {
    return is_subsumed(a, b->right);
  } else if (a->type == Resource) {
    if (eq(a->value, b->value)) return true;
    else return is_subsumed(a, b->right);
  } else if (a->type == Simple) {
    if (eq(a->value, b->value) &&
	(hasprefix(a->lang, b->lang) || eq(b->lang, "x-default"))) return true;
    else return is_subsumed(a, b->right);
  } else {			/* TODO: check complex values */
    return false;
  }
}


/* simplify -- remove redundant values from an Alt or Bag */
static void simplify(Store *store)
{
  Store *p = store;

  while (*p)
    if (is_subsumed(*p, *store)) delete_node(p);
    else p = &(*p)->right;
}


/* store_simplify -- simplify Alt, Bag and simple qualified properties */
EXPORT void store_simplify(Store *store)
{
  if (*store) {
    store_simplify(&(*store)->left);  /* Simplify left branch */
    store_simplify(&(*store)->right); /* Simplify right bracnh */

    if ((*store)->type != Simple && (*store)->type != Resource)
      simplify(&(*store)->sub);

    if ((*store)->type != Simple && (*store)->type != Resource &&
	(*store)->sub == NULL) {
      delete_node(store);  /* Delete empty array or structure */
    }
  }
}
