/*
 * xmptool - manipulate XMP files
 *
 * An XMP file is a file with RDF data in the RDF subset defined by
 * Adobe XMP.
 *
 * XMP is normally embedded in other files (such as JPEG, TIFF or PDF
 * files) to add metadata to them. This program does *not* handle
 * embedded XMP. Use a program such as rdjpgxmp or rdjpgapp to extract
 * the XMP from JPEG/JFIF and wrjpgapp to put it back in. Several
 * Adobe programs (such as Adobe Illustrator) can also extract XMP
 * from image formats supported by Adobe into a separate file and copy
 * XMP from an external file into an image file.
 *
 * TODO: A way to delete a sub-property of a structured value or the
 * nth value of a Seq.
 *
 * TODO: Exit codes from sysexits(3) instead of 0 or 1.
 *
 * TODO: Free all memory on the heap.
 *
 * Author: Bert Bos
 * Created: 16 August 2018
 *
 * Copyright © 2018 World Wide Web Consortium, (Massachusetts
 * Institute of Technology, European Research Consortium for
 * Informatics and Mathematics, Keio University, Beihang). All Rights
 * Reserved. See http://www.w3.org/Consortium/Legal/copyright-software
 */
#include "config.h"
#include <stdio.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <ctype.h>
#if STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
#endif
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include <stdbool.h>
#include <err.h>
#include "export.h"
#include "heap.e"
#include "types.e"
#include "html.e"
#include "scan.e"
#include "tree.e"
#include "propstore.e"
#include "printxmp.e"
#include "namespaces.e"
#include "extract.e"
#include "xmlnames.e"
#include "printsh.e"

static Tree tree;


/* add_utf8 -- add Unicode character n encoded as UTF-8 to string s */
static void add_utf8(string *s, unsigned long n)
{
  if (n <= 0x7F) {
    *(*s)++ = (char)n;
  } else if (n <= 0x7FF) {
    *(*s)++ = (char)(0xC0 | (n >> 6));
    *(*s)++ = (char)(0x80 | (n & 0x3F));
  } else if (n <= 0xFFFF) {
    *(*s)++ = (char)(0xE0 | (n >> 12));
    *(*s)++ = (char)(0x80 | ((n >> 6) & 0x3F));
    *(*s)++ = (char)(0x80 | (n & 0x3F));
  } else if (n <= 0x1FFFFF) {
    *(*s)++ = (char)(0xF0 | (n >> 18));
    *(*s)++ = (char)(0x80 | ((n >> 12) & 0x3F));
    *(*s)++ = (char)(0x80 | ((n >> 6) & 0x3F));
    *(*s)++ = (char)(0x80 | (n & 0x3F));
  } else if (n <= 0x3FFFFFF) {
    *(*s)++ = (char)(0xF0 | (n >> 24));
    *(*s)++ = (char)(0x80 | ((n >> 18) & 0x3F));
    *(*s)++ = (char)(0x80 | ((n >> 12) & 0x3F));
    *(*s)++ = (char)(0x80 | ((n >> 6) & 0x3F));
    *(*s)++ = (char)(0x80 | (n & 0x3F));
  } else if (n <= 0x7FFFFFFF) {
    *(*s)++ = (char)(0xF0 | (n >> 30));
    *(*s)++ = (char)(0x80 | ((n >> 24) & 0x3F));
    *(*s)++ = (char)(0x80 | ((n >> 18) & 0x3F));
    *(*s)++ = (char)(0x80 | ((n >> 12) & 0x3F));
    *(*s)++ = (char)(0x80 | ((n >> 6) & 0x3F));
    *(*s)++ = (char)(0x80 | (n & 0x3F));
  } else {
    errx(1, "Impossible numerical entity: &#x%lx;", n);
  }
}


/* unescape -- replace XML character entities */
static void unescape(const string s)
{
  string p, q;

  /* TODO: Warn for incorrect entities. */
  for (p = q = s; *p;)
    if (*p != '&') {*q++ = *p++;}
    else if (hasprefix(++p, "amp;")) {*q++ = '&'; p += 4;}
    else if (hasprefix(p, "lt;")) {*q++ = '<'; p += 3;}
    else if (hasprefix(p, "gt;")) {*q++ = '>'; p += 3;}
    else if (hasprefix(p, "quot;")) {*q++ = '"'; p += 5;}
    else if (hasprefix(p, "apos;")) {*q++ = '\''; p += 5;}
    else if (*p != '#') {*q++ = '&';} /* Unknown entity. Just copy. */
    else if (*(p+1) == 'x') {add_utf8(&q,strtoul(p+2,&p,16)); if (*p==';') p++;}
    else {add_utf8(&q, strtoul(p + 1, &p, 10)); if (*p == ';') p++;}
  *q = '\0';
}

/* handle_error -- called when a parse error occurred */
void handle_error(void *clientdata, const string s, int lineno)
{
  warnx("%d: %s", lineno, s);
}


/* start -- called before the first event is reported */
void* start(void)
{
  tree = create();
  return NULL;
}
  

/* handle_text -- called after a text chunk is parsed */
void handle_text(void *clientdata, string text)
{
  unescape(text);
  tree = tree_append_text(tree, text);
}


/* handle_starttag -- called after a start tag is parsed */
void handle_starttag(void *clientdata, string name, pairlist attribs)
{
  pairlist a;

  for (a = attribs; a; a = a->next) unescape(a->value);
  tree = tree_push(tree, name, attribs);
  free(name);
}


/* handle_emptytag -- called after an empty tag is parsed */
void handle_emptytag(void *clientdata, string name, pairlist attribs)
{
  pairlist a;

  for (a = attribs; a; a = a->next) unescape(a->value);
  tree = tree_push(tree, name, attribs);
  tree = tree_pop(tree, name);
  free(name);
}


/* handle_endtag -- called after an endtag is parsed (name may be "") */
void handle_endtag(void *clientdata, string name)
{
  tree = tree_pop(tree, name);
  free(name);
}


/* expand_prefix -- replace "foo:bar" by "URIbar" */
static void expand_prefix(SymbolTable prefixes, string *t)
{
  string prefix, name, s, p;

  if (!(p = strchr(*t, ':'))) {prefix = ""; name = *t;}
  else {*p = '\0'; prefix = *t; name = p + 1;}

  if (!(s = globalname(prefixes, prefix, name)))
    errx(1, "No definition found for namespace prefix \"%s\"\n", prefix);

  dispose(*t);
  *t = s;
}


/* expand_prefixes -- walk through the document to expand namespace prefixes */
void expand_prefixes(Tree t, SymbolTable prefixes)
{
  pairlist *a, b;
  Tree h;

  assert(t->tp == Root || t->tp == Element || t->tp == Text);

  if (t->tp == Element) {

    /* Extract prefix definitions from attributes and store them. */
    a = &t->attribs;
    while (*a)
      if (eq((*a)->name, "xmlns")) {
	symboltable_add(&prefixes, "", (*a)->value);
	*a = (*a)->next;		/* Remove the attribute */
      } else if (hasprefix((*a)->name, "xmlns:")) {
	symboltable_add(&prefixes, (*a)->name + 6, (*a)->value);
	*a = (*a)->next;		/* Remove the attribute */
      } else {
	a = &(*a)->next;		/* Go to next attribute */
      }

    /* Replace prefixes by namespace URI */
    expand_prefix(prefixes, &t->name);
    for (b = t->attribs; b; b = b->next) expand_prefix(prefixes, &b->name);
  }

  /* Recursively process child elements */
  for (h = t->children; h; h = h->sister) expand_prefixes(h, prefixes);
}


/* print_cb -- callback to print a property value */
static void print_cb(string name, PropertyType type, string value, Store sub,
		     string lang, void *data)
{
  string t, s = (string)data;

  /* If verbose, concatenate parent property (if any) and currrent name */
  if (!s) t = NULL;		/* Not verbose */
  else if (!name || !*name || eq(name, VALUE)) t = newstring(s);
  else if (!*s) t = newstring(name); /* No parent to concatenate with */
  else t = strapp(NULL, s, " ", name, NULL); /* Concatenate */

  /* Add language */
  if (s && lang && *lang && !eq(lang,"x-default"))
    strapp(&t, "[", lang, "]", NULL);

  /* Print any text value */
  if (type == Simple || type == Resource) {
    if (t) printf("%s\t", t);	/* Verbose */
    printf("%s\n", value);
  }
  
  /* Print any array items, fields or qualifiers */
  store_walk(sub, print_cb, t);
  dispose(t);
}


/* print_property -- look up a property and print it */
static void print_property(Store store, string name, string lang, bool verbose)
{
  (void) store_lookup(store, name, lang, print_cb, verbose ? "" : NULL);
}


/* usage -- print usage message and exit */
static void usage(int exitcode, string prog)
{
  fprintf(exitcode ? stderr : stdout, "%1$s version %2$s\n\
Usage: %1$s [-v] [-l language] property [file]\n\
       %1$s -w [-r] [-l language] property value [file]\n\
       %1$s -d [-l language] property [file]\n\
       %1$s -c [file]\n\
       %1$s -s [file]\n",
	  prog, VERSION);
  exit(exitcode);
}


/* main -- main body */
int main(int argc, char *argv[])
{
  enum {PRINT, WRITE, DELETE, SHELL, CLEAN} task = PRINT;
  string language = NULL, property = NULL, name, value = NULL, p, q;
  PropertyType type = Simple;
  SymbolTable prefixes;
  bool verbose = false;
  FILE *input;
  Store store;
  int c;

  /* Bind the parser callback routines to our handlers */
  set_error_handler(handle_error);
  set_start_handler(start);
  set_text_handler(handle_text);
  set_starttag_handler(handle_starttag);
  set_emptytag_handler(handle_emptytag);
  set_endtag_handler(handle_endtag);

  /* Parse command line */
  while ((c = getopt(argc, argv, ":cvrwdl:sh")) != -1)
    switch (c) {
    case 'w': if (task == PRINT) task = WRITE; else usage(1, argv[0]); break;
    case 'l': language = optarg; break;
    case 'r': type = Resource; break;
    case 'd': if (task == PRINT) task = DELETE; else usage(1, argv[0]); break;
    case 's': if (task == PRINT) task = SHELL; else usage(1, argv[0]); break;
    case 'v': if (task == PRINT) verbose = true; else usage(1, argv[0]); break;
    case 'c': if (task == PRINT) task = CLEAN; else usage(1, argv[0]); break;
    case 'h': usage(0, argv[0]); break;
    default: usage(1, argv[0]);
    }
  if (task != SHELL && task != CLEAN) {
    if (argc == optind) usage(1, argv[0]);
    property = argv[optind++];
  }
  if (task == WRITE) {
    if (argc != optind) value = argv[optind++]; else usage(1, argv[0]);
  }
  if (optind == argc) set_yyin(stdin, "<stdin>");
  else if (optind != argc - 1) usage(1, argv[0]);
  else if ((input = fopen(argv[optind], "r"))) set_yyin(input, argv[optind]);
  else err(2, "%s", argv[optind]);

  /* Step 1: parse input and build a document tree. */
  if (yyparse() != 0) exit(3);

  /* Step 2: expand namespace prefixes. */
  tree = get_root(tree);
  prefixes = symboltable_create();
  expand_prefixes(tree, prefixes);

  /* Step 3: convert the RDF/XML tree to a property store. */
  store = store_create();
  extract_properties(tree, &store);

  /* Step 4: perform requested operation. */
  switch (task) {
  case WRITE:
    assert(property);
    assert(value);
    /* Check syntax: Set name to point after the last non-name character */
    for (name = p = q = property; *p; p = q)
      if (!is_name_char(p, &q)) name = q;
    if (!is_name_start_char(name, NULL)) /* Invalid start char or empty name */
      errx(1, "Property name must end with a valid identifier");
    if (!language) language = ""; /* Empty language */
    store_add(&store, property, type, value, NULL, language);
    print_xmp(store);
    break;
  case SHELL:
    print_sh(store);
    break;
  case PRINT:
    assert(property);
    if (!language) language = "*"; /* Match any language */
    print_property(store, property, language, verbose);
    break;
  case DELETE:
    assert(property);
    if (!language) language = "*"; /* Match any language */
    store_delete(&store, property, language);
    print_xmp(store);
    break;
  case CLEAN:
    store_simplify(&store);
    print_xmp(store);
    break;
  default:
    assert(!"Cannot happen!");
  }

  return 0;
}
